import os
import glob
import re
from .tempfiles import create_temp_file

from ..utils import import_module
ffmpeg = import_module('ffmpeg', 'ffmpeg-python')


valid_frames_exts = ('png', 'jpg', 'jpeg', 'bmp', 'tiff', 'tif')


def get_frames_from_directory(directory, exts=valid_frames_exts, sort=True):
	"""Return a list of all paths to files in directory which have a valid frame extension"""
	files = [os.path.join(directory, f) for f in os.listdir(directory) if os.path.splitext(f)[-1][1:] in exts]
	if sort:
		files = sorted(files)
	return files


def ffmpeg_input_from_files(file_list, frame_rate=25):
	"""Return ffmpeg input from list of files"""
	file_list_loc = create_temp_file(ext='.txt')

	with open(file_list_loc, 'w') as f:
		for file in file_list:
			abs_file = os.path.abspath(file)
			f.write(f"file '{abs_file}'\n")

	return ffmpeg.input(file_list_loc, r=str(frame_rate), f='concat', safe='0')


def frames_to_video(*args, frame_list=None, frame_fmt=None, directory=None,
					frame_rate=30, output_loc='./output.mp4', output_fmt='mp4', pix_fmt='yuv420p',
					overwrite=True, loglevel='error',
					delete_images=False):
	"""
	Need to input one of frame_list or frame_fmt or directory

	:param frame_list: [optional] list of frames to concatenate into a video
	:param frame_fmt: [optional] frame format to use as input
	:param directory: [optional] directory to use as input (loads sorted images in directory)

	:param frame_rate: frame rate of video
	:param output_loc: video file location
	:param output_fmt: video file format
	:param pix_fmt: pixel format (see ffmpeg documentation, default yuv420p)

	:param overwrite: overwrite existing video file
	:param loglevel: ffmpeg loglevel (see ffmpeg documentation, default error)

	:param delete_images: delete images after video is created
	:return:
	"""

	assert args == (), 'frames_to_video() takes keyword arguments only'

	# Assert exactly 1 of frame_list, frame_fmt, directory is provided
	assert sum([frame_list is not None, frame_fmt is not None,
				directory is not None]) == 1, 'Exactly one of frame_list, frame_fmt, directory must be provided'

	vid_input = None
	file_list = []
	if frame_list:
		assert frame_list, 'No frames provided'
		vid_input = ffmpeg_input_from_files(frame_list, frame_rate=frame_rate)
		file_list = frame_list

	elif frame_fmt:
		vid_input = ffmpeg.input(frame_fmt)

		# Get file list using glob (note: this does not match the exact frame format, so may not be fully robust)
		file_list = glob.glob(re.sub(r'%\d*[df]', '*', frame_fmt))

	elif directory:
		file_list = get_frames_from_directory(directory)
		assert file_list, 'No valid frames found in directory'

		vid_input = ffmpeg_input_from_files(file_list, frame_rate=frame_rate)

	(vid_input
		.output(output_loc, format=output_fmt, r=frame_rate, pix_fmt=pix_fmt)
		.global_args('-loglevel', loglevel)
		.run(overwrite_output=overwrite)
		)

	if delete_images:
		for f in file_list:
			os.remove(f)

	print('Video created at ' + output_loc)



from .node_group import CompositorNodeGroup
import bpy

class DepthVis(CompositorNodeGroup):
	def __init__(self, node_tree, max_depth=1, col=(1, 1, 1)):
		super().__init__(name='DepthVis', node_tree=node_tree)
		self.max_depth = 1

		self.group.inputs.new('NodeSocketFloat', 'Depth')
		self.group.outputs.new('NodeSocketColor', 'Image')

		self.map_range_node = self.group.nodes.new('CompositorNodeMapRange')
		self.map_range_node.inputs['From Max'].default_value = max_depth

		self.rgb_node = self.group.nodes.new('CompositorNodeRGB')
		if len(col) == 3:
			col = (*col, 1)

		self.rgb_node.outputs['RGBA'].default_value = col

		self.multiple_node = self.group.nodes.new('CompositorNodeMath')
		self.multiple_node.operation = 'MULTIPLY'

		self.group.links.new(self.input_node.outputs['Depth'], self.map_range_node.inputs['Value'])
		self.group.links.new(self.map_range_node.outputs['Value'], self.multiple_node.inputs[0])
		self.group.links.new(self.rgb_node.outputs['RGBA'], self.multiple_node.inputs[1])
		self.group.links.new(self.multiple_node.outputs['Value'], self.output_node.inputs['Image'])


"""Context managers and operations for handling Blender"""
import inspect

import bpy
import numpy as np
import mathutils
from typing import Union, List, Tuple

# define Blender Array type
blender_array_type = Union[List[float], mathutils.Vector, np.ndarray, Tuple[float, ...], Tuple[int, ...]]
blender_array_or_scalar = Union[blender_array_type, int, float]

def _euler_from(a: mathutils.Euler, b: mathutils.Euler):
	"""Get euler rotation from a to b"""
	return (b.to_matrix() @ a.to_matrix().inverted()).to_euler()


def _euler_add(a: mathutils.Euler, b: mathutils.Euler):
	"""Compute euler rotation of a, followed by b"""
	return (a.to_matrix() @ b.to_matrix()).to_euler()


class GetNewObject():
	"""Context manager for getting the newly imported object(s) to the scene.

	On exit, will return the newly imported object(s).

	Assumes that either (1) only one object is imported, or (2) there is a hierarchy to the imported objects, and the top level object is the one to return."""

	def __init__(self, scene):
		self.scene = scene
		self.imported_obj = None

	def __enter__(self):
		self.old_objs = set(self.scene.objects)

	def __exit__(self, *args):
		x = set(self.scene.objects) - self.old_objs
		assert len(x) > 0, "No imported objects found..."

		if len(x) == 1:
			self.imported_obj = x.pop()

		else:

			# assume there is a hierarchy to the objects - get the highest
			parent_obj = None
			for obj in x:
				if obj.parent is None:
					parent_obj = obj
					break

			assert parent_obj is not None, "Multiple objects loaded,  but no parent object found..."
			self.imported_obj = parent_obj


class SelectObjects:
	"""Context manager for selecting objects.
	On exit, will reselect the objects that were selected before entering the context."""

	def __init__(self, objects: list = ()):
		"""Initialize with a list of objects to select
		:param objects: list of bpy.types.Object"""
		self.objects = objects

	def __enter__(self):
		self.old_objs = bpy.context.selected_objects
		# deselect all
		bpy.ops.object.select_all(action='DESELECT')

		# select objects
		for obj in self.objects:
			obj.select_set(True)

	def __exit__(self, *args):
		for obj in self.objects:
			obj.select_set(False)

		for obj in self.old_objs:
			obj.select_set(True)


class SetMode:
	"""Context manager for changing the mode of a specific object in Blender (e.g., to 'POSE'),
	returning to the original mode on exit."""

	def __init__(self, target_mode:str, object:bpy.types.Object=None):
		"""Initialize with the target mode and object
		:param target_mode: Mode to set the object to
		:param object: bpy.types.Object to set the mode of"""
		self.target_mode = target_mode.upper()
		self.original_mode = None
		self.obj = object
		self.original_active_object = None

	def __enter__(self):
		self.original_active_object = bpy.context.view_layer.objects.active

		if self.obj:
			bpy.context.view_layer.objects.active = self.obj

		self.original_mode = bpy.context.object.mode
		bpy.ops.object.mode_set(mode=self.target_mode)

	def __exit__(self, type, value, traceback):
		bpy.ops.object.mode_set(mode=self.original_mode)
		bpy.context.view_layer.objects.active = self.original_active_object


def get_node_by_name(node_tree: bpy.types.NodeTree, key: str, raise_error=False) -> bpy.types.Node:
	"""Given a nodetree and a key, return the first node found with label matching key.

	:param node_tree: Node tree to search
	:param key: Key to search for
	:param raise_error: If True, raise KeyError if key not found
	:return: Node with matching label"""
	for node in node_tree.nodes:
		if node.name == key:
			return node

	if raise_error:
		raise KeyError(f"Key {key} not found in node tree!\nLabels are: {[n.name for n in node_tree.nodes]}")


def handle_vec(vec, expected_length: int = 3) -> mathutils.Vector:
	"""Check `vec` is expected_length. Convert from tuple or ndarray to mathutils.Vector.

	:param vec: Vector to check
	:param expected_length: Expected length of vector
	"""

	if isinstance(vec, (tuple, list)):
		vec = mathutils.Vector(vec)
	elif isinstance(vec, np.ndarray):
		vec = mathutils.Vector(vec.tolist())

	if len(vec) != expected_length:
		raise ValueError("Vector must be length {}".format(expected_length))

	return vec


def animatable_property(data_path: str, use_data_object: bool = False) -> callable:
	"""Decorator that wraps around a function to take a frame number and value, and set the property at that frame.

	example usage::

		@animatable('location')
		def set_location(self, value):
			self._location = value

	If you want to set the property at the current frame, use the setter as normal:

	``obj.set_location((1, 2, 3))``

	To set the property at a specific frame, use the decorator:

	``obj.set_location((1, 2, 3), frame=10)``

	Which will call the set_location function, followed by

	``self.object.keyframe_insert(data_path='location', frame=10)``

	:param data_path: the data path of the property to set
	:param use_data_object: whether to use the data object or the object itself
	"""

	def wrapper(func):

		original_sig = inspect.signature(func)
		original_params = list(original_sig.parameters.values())

		def subwrapper(self: 'BsynObject', value, *args, frame=None, **kwargs):
			frame = args[0] if len(args) > 0 else frame
			func(self, value, **kwargs)
			if frame is not None:
				object = self.object if not use_data_object else self.object.data
				object.keyframe_insert(data_path=data_path, frame=frame)


		# store original parameters here - we need them for type hinting
		param_types = {param.name: param.annotation for param in original_params}

		# remove type hints from signature as they will be added to the docstring. Add frame = None
		new_params = \
			[inspect.Parameter(param.name, inspect.Parameter.POSITIONAL_OR_KEYWORD, default=param.default) for param in original_params] + \
			[inspect.Parameter("frame", inspect.Parameter.POSITIONAL_OR_KEYWORD, default=None)]

		subwrapper.__signature__ = original_sig.replace(parameters=new_params)

		# We need to make the following updates to the docstring:
		# Add parameter 'frame' (matching indent)
		# Copy type hints from signature (to overcome the lack of sphinx autodoc for runtime docstrings)
		doc_lines = []
		line_gen = (line for line in (func.__doc__ or '').split("\n")) # generator of all lines
		indentation = ''
		started_params = False
		for line in line_gen:
			line_starts_with_params = line.lstrip().startswith(":param")
			doc_lines.append(line)
			if not line_starts_with_params and started_params: # end of param block
				break

			if line_starts_with_params:
				indentation = line[:line.index(":param")]
				name = line.split(':')[1].strip().removeprefix('param ')
				param_type = param_types.get(name, inspect._empty)
				if param_type is not inspect._empty:  # If the parameter has a type hint
					if hasattr(param_type, '__name__'):
						class_name = param_type.__name__

					else: # get class name e.g. blendersynth.utils.Object
						class_name = str(param_type)

					doc_lines.append(f'{indentation}:type {name}: :class:`~{class_name}`')

		# add frame param
		doc_lines.append(f"\n{indentation}:param frame: Optional frame for animating \n{indentation}:type frame: :class:`~int`")

		# add rest of lines
		for line in line_gen:
			doc_lines.append(line)

		subwrapper.__doc__ = "\n".join(doc_lines)

		return subwrapper

	return wrapper



import bpy
from .utils import project_points
import mathutils
import numpy as np
from ..blender.mesh import Mesh
from typing import List

def get_axes(object: Mesh, camera: bpy.types.Camera = None,
				 scene:bpy.types.Scene=None,
				 normalized=True,
				 invert_y=True) -> tuple:
	"""Get the unit axes of an object in camera space.

	Return in (4 x 2) centroid + 3 pixel positions of ends of unit axes [XYZ] in XY

	:param object: Mesh
	:param camera: bpy.types.Camera (if None, use bpy.context.scene.camera)
	:param scene: bpy.types.Scene (if None, use bpy.context.scene)
	:param normalized: if True, return normalized coordinates (0-1) instead of pixel coordinates
	:param invert_y: if True, y is measured from the top of the image, otherwise from the bottom (Blender measures from bottom)
	"""

	if camera is None:
		camera = bpy.context.scene.camera

	if scene is None:
		scene = bpy.context.scene

	if len(object._meshes) > 1:
		raise NotImplementedError('Axes only supported for single meshes')

	centroid = np.array(object.centroid())
	verts = np.concatenate([centroid[None, :], centroid[None, :] + object.axes], axis=0)
	coords_2d = project_points(verts, scene, camera, invert_y=invert_y)

	if normalized:
		coords_2d[:, 0] /= scene.render.resolution_x
		coords_2d[:, 1] /= scene.render.resolution_y

	return coords_2d

def get_multiple_axes(objects: List[Mesh], camera: bpy.types.Camera = None,
				 scene:bpy.types.Scene=None,
				 normalized:bool=False, invert_y:bool=True) -> List[tuple]:
	"""Run `get_axes` for multiple objects

	:param objects: List of Mesh objects
	:param camera: bpy.types.Camera (if None, use bpy.context.scene.camera)
	:param scene: bpy.types.Scene (if None, use bpy.context.scene)
	:param normalized: if True, return normalized coordinates (0-1) instead of pixel coordinates
	:param invert_y: if True, y is measured from the top of the image, otherwise from the bottom (Blender measures from bottom)
	"""

	return [get_axes(obj, camera=camera, scene=scene,
						   normalized=normalized, invert_y=invert_y) for obj in objects]


from typing import Union, List, Tuple
from mathutils import Vector
import numpy as np

types_loc = 'blendersynth.utils.types'

VectorLikeAlias = f'{types_loc}.VectorLike'
VectorLike = Union[Vector, np.ndarray, List, Tuple]
"""A type hint that represents vector-like objects. Can be any of:

- :class:`mathutils.Vector`
- :class:`numpy.ndarray`
- :class:`list`
- :class:`tuple`
"""

VectorLikeOrScalarAlias = f'{types_loc}.VectorLikeOrScalar'
VectorLikeOrScalar = Union[VectorLike, int, float]
"""A type hint that represents vector-like objects. Can be any of:

- :class:`~VectorLike`
- :class:`int`
- :class:`float`
"""


import tempfile
import os

tempdir = os.path.join(tempfile._get_default_tempdir(), 'blendersynth')
os.makedirs(tempdir, exist_ok=True)


def create_temp_file(ext='.png'):
	"""Create and return a temporary filename that will be in the user's tempdir.
	These can all be cleaned up after use by the `cleanup_temp_files` function.

	:param ext: file extension to use"""

	if not ext.startswith('.'):
		ext = '.' + ext

	return tempfile.mktemp(dir=tempdir) + ext


def cleanup_temp_files():
	"""Delete all temporary files in the `blendersynth` tempdir"""
	for f in os.listdir(tempdir):
		os.remove(os.path.join(tempdir, f))


import subprocess
import os
from blendersynth.utils.blender_setup.blender_locator import find_blender_python, get_blender_path, remove_config, write_to_config, read_from_config, remove_from_config

dependencies = ['imageio', 'numpy', 'appdirs', 'tqdm', 'opencv-python', 'ffmpeg-python']

def check_module(python_executable, module_name):
	try:
		subprocess.check_call([python_executable, '-m', 'pip', 'show', module_name],
							  stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)
		return True
	except subprocess.CalledProcessError:
		return False

def install_module(python_executable, module_name, is_test_pypi=False,
				   version=None, upgrade=True, editable=False):

	commands = [python_executable, '-m', 'pip']
	commands += ['install']
	if is_test_pypi:
		commands += ['-i', 'https://test.pypi.org/simple/']

	if editable:
		commands += ['-e']

	commands += [module_name + (f"=={version}" if version is not None else "")]
	if upgrade:
		commands += ['--upgrade']

	try:
		subprocess.check_call(commands)
	except subprocess.CalledProcessError as e:
		raise Exception(f"Could not install {module_name} via pip. Error: {e}")

def check_blender_install(force_all=False,
						  force_find_blender=False,
						  force_find_blender_python=False,
						  force_install_dependencies=False,
						  blendersynth_from_local=False,
						  blendersynth_editable=False):
	"""Check if Blender is installed correctly and has all necessary packages.
	If not, run first time setup.

	On first time setup, will create a file, config.ini, in the user's config,
	containing the necessary info.

	Force: if True, will run first time setup (overwriting any existing config.ini)
	regardless"""

	if force_all:
		remove_config() # remove config file if it exists to force first time setup

	if force_find_blender: remove_from_config('BLENDER_PATH')
	if force_find_blender_python: remove_from_config('BLENDER_PYTHON_PATH')
	if force_install_dependencies: remove_from_config('DEPENDENCIES_INSTALLED')

	blender_path = get_blender_path()
	python_path = find_blender_python(blender_path)

	if not read_from_config('DEPENDENCIES_INSTALLED') == 'True':
		# check if blender's python has all necessary packages
		for dependency in dependencies:
			if not check_module(python_path, dependency):
				install_module(python_path, dependency)

		# First uninstall blendersynth if present (in case of updates)
		if check_module(python_path, 'blendersynth'):
			subprocess.check_call([python_path, '-m', 'pip', 'uninstall', '-y', 'blendersynth'])

		# Install blendersynth package to blender's python
		if blendersynth_from_local:
			# Install from local setup.py
			setup_py_loc = os.path.join(os.path.dirname(__file__), '..', '..', '..', 'setup.py')
			directory_loc = os.path.dirname(setup_py_loc)
			install_module(python_path, directory_loc, editable=blendersynth_editable)

		else:
			# Install from pypi
			install_module(python_path, 'blendersynth', upgrade=True)

		write_to_config('DEPENDENCIES_INSTALLED', 'True')


"""Quick utility to run the current script from Blender"""
import subprocess
import inspect
import sys
from ..utils.blender_setup.blender_locator import get_blender_path
from ..file.tempfiles import cleanup_temp_files as cleanup

def run_this_script(debug:bool=False):
	"""Run the script in which this function is called from Blender

	:param debug: If True, open a Blender instance after all code is executed, otherwise run in background"""
	blender_path = get_blender_path()
	if blender_path != sys.argv[0]:  # if blender is not running this script

		caller_path = inspect.stack()[1].filename # path of script that called this function

		commands = [blender_path] + \
			['--background'] * (not debug) + \
			['--python', caller_path]

		subprocess.call(commands)

		cleanup()  # cleanup temp files
		sys.exit()  # exit the script once blender is finished


# First, make sure Blender has been set up properly
import sys
from .utils.blender_setup import check_blender_install

building_docs = 'sphinx' in sys.modules

if not building_docs:
	check_blender_install(blendersynth_from_local='--local' in sys.argv) # check install here

def fix_blender_install(local=False, editable=False):
	check_blender_install(force_all=True, blendersynth_from_local=local, blendersynth_editable=editable)

def fix_blender_modules(local=False, editable=False):
	check_blender_install(force_install_dependencies=True, blendersynth_from_local=local,
						  blendersynth_editable=editable)

from .run.run_this_script import run_this_script

import sys
from .utils.blender_setup.blender_locator import get_blender_path

if building_docs or get_blender_path() == sys.argv[0]:  # if blender is running this script, or if building docs
	import bpy
	from bpy import *
	from .blender.mesh import Mesh
	from .blender.material import Material
	from .blender.curve import Curve
	from .blender import render
	from .blender.compositor.compositor import Compositor
	from .blender import aov
	from .file.dataset_inputs import Inputs, DebugInputs
	from . import file
	from .run.blender_interface import log_event
	from .blender.world import world
	from .blender.light import Light
	from .blender.camera import Camera
	from . import annotations
	from .file.tempfiles import cleanup_temp_files as cleanup

	# set render engine to cycles
	render.set_engine('CYCLES')

	# common aliases
	def load_blend(src):
		return bpy.ops.wm.open_mainfile(filepath=src)

	# Clear default cube and light
	import bpy
	bpy.data.objects['Cube'].select_set(True)
	bpy.data.objects['Light'].select_set(True)
	bpy.ops.object.delete()

else:
	# Imports for non blender
	from .run.run import execute_jobs
	from . import file



import ffmpeg
import os

def hstack(videos:list, out_path:str, overwrite:bool=True, loglevel:str='error'):
	"""Horizontally stack videos.

	:param videos: list of video paths
	:param out_path: output video path
	:param overwrite: overwrite existing video file
	:param loglevel: ffmpeg loglevel (see ffmpeg documentation, default error)"""

	(ffmpeg.filter([ffmpeg.input(v) for v in videos], 'hstack')
	 .output(out_path)
	 .global_args('-loglevel', loglevel)
	 .run(overwrite_output=overwrite)
	 )


import bpy
from ..utils.node_arranger import tidy_tree
from typing import Union

class World():
	"""World object - for managing world nodes, HDRIs, etc."""

	def __init__(self):
		self.mode = 'Color'
		self.world = bpy.data.worlds["World"]
		self.node_tree = self.world.node_tree
		self.world_nodes = self.node_tree.nodes

		self.hdri_link = None
		self._setup_nodes()

		self.set_color((0.051,) * 3)  # match blender's default world colour

	def _check_exists(self):
		"""Node tree may have been wiped, eg by loading a new file.
		Check if the nodes exist, and if not, recreate them."""

		try:
			self.node_tree.rna_type
		except ReferenceError:
			self.__init__()

	def _setup_nodes(self):

		self._check_exists()
		nodes = self.world_nodes
		nodes.clear()

		self.node_texture = nodes.new(type='ShaderNodeTexEnvironment')
		self.node_background = nodes.new(type='ShaderNodeBackground')
		self.node_output = nodes.new(type='ShaderNodeOutputWorld')

		self.node_tree.links.new(self.node_background.outputs["Background"], self.node_output.inputs["Surface"])

		tidy_tree(self.node_tree)

	def _setup_color(self):

		self._check_exists()
		if self.mode == 'Color':
			return

		if self.hdri_link is not None:
			self.node_tree.links.remove(self.hdri_link)

		tidy_tree(self.node_tree)

		self.mode = 'Color'

	def _setup_hdri(self):

		self._check_exists()
		if self.mode == 'HDRI':
			return

		# Link the nodes
		self.hdri_link = self.node_tree.links.new(self.node_texture.outputs["Color"],
												  self.node_background.inputs["Color"])
		tidy_tree(self.node_tree)

		self.mode = 'HDRI'

	def set_color(self, color: Union[list, tuple], affect_scene:bool=True):
		"""Set the world color.

		:param color: RGB or RGBA color
		:param affect_scene: Toggle for whether color's lighting should affect the scene (if False, functions as a solid background color)"""

		self._setup_color()

		assert len(color) in [3, 4], "Color must be RGB or RGBA"

		if len(color) == 3:
			color = (*color, 1.0)

		self.node_background.inputs["Color"].default_value = color
		self._lighting_from_background(affect_scene)

	def set_hdri(self, pth:str, affect_scene:bool=True, intensity:float=None):
		"""Set the HDRI image location

		:param pth: Path to the HDRI image (.hdr or .exr)
		:param affect_scene: Toggle for whether color's lighting should affect the scene (if False, functions as a solid background color)
		:param intensity: [Optional] Set the intensity of the HDRI lighting"""


		self._setup_hdri()
		self.world_nodes['Environment Texture'].image = bpy.data.images.load(pth)
		self._lighting_from_background(affect_scene)

		if intensity:
			self.set_intensity(intensity)

	def set_intensity(self, intensity:float=1.):
		"""Set the intensity of the color/HDRI.

		:param float: The intensity value
		"""
		self.world_nodes["Background"].inputs[1].default_value = intensity

	def set_transparent(self, transparent=True):
		bpy.context.scene.render.film_transparent = transparent

	def _lighting_from_background(self, val=True):
		"""Change the ability for the background to influence lighting on the scene"""
		self.world.cycles_visibility.diffuse = val


world = World()



"""Shader AOV manager"""

import bpy
from ..utils.node_arranger import tidy_tree

ref_frames = ['CAMERA', 'WORLD', 'OBJECT']

# Acceptable socket types for AOV colors & nodes
_socket_color_types = (bpy.types.NodeSocketVector, bpy.types.NodeSocketColor)
_socket_value_types = (bpy.types.NodeSocketFloat, bpy.types.NodeSocketInt)

class AOV:
	"""A generic Arbitrary Output Value.
	An AOV is a float or color value that can be output from a shader to the renderer.
	See `Blender docs <https://docs.blender.org/manual/en/latest/render/shader_nodes/output/aov.html>`_ for more info.
	"""

	AOV_TYPE = 'COLOR'
	def __init__(self, name=None, **kwargs):
		if name is None:
			name = self.__class__.__name__

		self.name = name
		self._aov = None
		self._add_to_view_layer()

	def _add_to_view_layer(self, view_layer=None):
		"""Add aov to view layer if not already there"""
		if view_layer is None:
			view_layer = bpy.context.view_layer

		for aov in view_layer.aovs:
			if aov.name == self.name:
				return

		aov = view_layer.aovs.add()
		aov.name = self.name
		self._aov = aov


	def add_to_shader(self, shader_node_tree):
		"""Add AOV to shader node tree"""
		# add_to_shader must return output socket to connect to node
		out_socket = self._add_to_shader(shader_node_tree)

		# check node socket is correct type
		AOV_TYPE = None
		if isinstance(out_socket, _socket_color_types):
			AOV_TYPE = 'COLOR'
		elif isinstance(out_socket, _socket_value_types):
			AOV_TYPE = 'VALUE'
		else:
			raise ValueError(f"Output of _add_to_layer must be in {_socket_color_types} if Color or {_socket_value_types} if value. Got: `{type(out_socket)}`")

		shader_aov_node = shader_node_tree.nodes.new('ShaderNodeOutputAOV')
		shader_aov_node.name = self.name
		shader_node_tree.links.new(out_socket, shader_aov_node.inputs[AOV_TYPE.title()])

		self._aov.type = self.AOV_TYPE
		tidy_tree(shader_node_tree)

	def _add_to_shader(self, shader_node_tree) -> bpy.types.NodeSocket:
		raise NotImplementedError

	def update(self):
		"""Some AOVs need an update before rendering (to change certain node properties)"""
		return

	def __str__(self):
		return self.name

class NormalsAOV(AOV):
	def __init__(self, name=None,
					ref_frame='CAMERA',
					order='XYZ',
					polarity=(1, 1, 1)):
		"""Given a shader node tree, add surface normals as output.
		:param shader_node_tree: Shader node tree to add AOV to
		:param aov_name: Name of AOV to add
		:param ref_frame: Reference frame to use for normals
		:param order: Order of components in RGB (default: XYZ)
		:param polarity: Polarity of XYZ (1 or -1)
		"""
		super().__init__(name)
		assert ref_frame in ref_frames, f"ref_frame must be one of {ref_frames}"

		self.ref_frame = ref_frame
		self.order = order
		self.polarity = polarity

	def _add_to_shader(self, shader_node_tree):

		geom_node = shader_node_tree.nodes.new('ShaderNodeNewGeometry')
		vec_transform = shader_node_tree.nodes.new('ShaderNodeVectorTransform')
		map_range_node = shader_node_tree.nodes.new('ShaderNodeMapRange')

		vec_transform.vector_type = 'NORMAL'
		vec_transform.convert_to = self.ref_frame

		# Set up mapping - with polarity
		map_range_node.data_type = 'FLOAT_VECTOR'
		for i in range(3):
			map_range_node.inputs[7].default_value[i] = - self.polarity[i]
			map_range_node.inputs[8].default_value[i] = self.polarity[i]

		# Set up ordering
		sep_xyz_node = shader_node_tree.nodes.new('ShaderNodeSeparateXYZ')
		comb_xyz_node = shader_node_tree.nodes.new('ShaderNodeCombineXYZ')
		for i in range(3):
			shader_node_tree.links.new(sep_xyz_node.outputs[self.order[i]], comb_xyz_node.inputs['XYZ'[i]])

		# Make necessary connections for shader graph
		shader_node_tree.links.new(geom_node.outputs['True Normal'], vec_transform.inputs['Vector'])
		shader_node_tree.links.new(vec_transform.outputs['Vector'], map_range_node.inputs['Vector'])
		shader_node_tree.links.new(map_range_node.outputs['Vector'], sep_xyz_node.inputs['Vector'])

		return comb_xyz_node.outputs['Vector']

	def update(self, scene=None):
		"""Some AOVs may need render_time updates from scene context, hence this method"""
		pass

class NOCAOV(AOV):
	"""Normalised object coordinates - Gives the position of the object in object space,
	normalised to the object's bounding box."""

	def _add_to_shader(self, shader_node_tree):
		texcon_node = shader_node_tree.nodes.new('ShaderNodeTexCoord')
		return texcon_node.outputs['Generated']

class UVAOV(AOV):
	"""UV coordinates"""

	def _add_to_shader(self, shader_node_tree):
		texcon_node = shader_node_tree.nodes.new('ShaderNodeTexCoord')
		return texcon_node.outputs['UV']


class AttrAOV(AOV):
	"""Take an object attribute, and output it as an AOV"""
	attribute_type = None
	attribute_name = None

	def _add_to_shader(self, shader_node_tree):
		attr_node = shader_node_tree.nodes.new('ShaderNodeAttribute')
		attr_node.attribute_type = self.attribute_type
		attr_node.attribute_name = self.attribute_name
		return attr_node.outputs['Instance Index']

class InstanceIDAOV(AttrAOV):
	"""Instance ID - given to each object on creation.
	Output is an integer corresponding to the object's instance ID (0-indexed)
	"""
	attribute_type = 'OBJECT'
	attribute_name = 'instance_id'

class ClassIDAOV(AttrAOV):
	"""Class ID - given to each object on creation.
	Output is an integer corresponding to the object's class ID (0-indexed)
	Class IDs can be manually set either when creating a Mesh, or by using the Mesh's set_class_id() method.
	If not set, will default to a different index from each primitive.
	"""
	attribute_type = 'OBJECT'
	attribute_name = 'class_id'

class AttrRGBAOV(AOV):
	"""
	For a given numerical attribute, outputs a color corresponding to the attribute's value.
	For N objects total, samples evenly on a distribution of hues, on a colour scale of S = 1, V = 1
	Runs update() method to change the value of N, which is called before rendering.
	N can be a property of the object to update.
	"""
	attribute_type = None
	attribute_name = None

	def __init__(self, name=None):
		super().__init__(name)

		# Create Int Index -> HSV as a node group, so the 'num_objects' parameter can be edited centrally
		self.group = bpy.data.node_groups.new(name='IdxToHue', type='ShaderNodeTree')

		self.group.inputs.new(f'NodeSocketInt', 'Index')
		self.group.outputs.new(f'NodeSocketColor', 'Color')

		self.input_node = self.group.nodes.new('NodeGroupInput')
		self.output_node = self.group.nodes.new('NodeGroupOutput')

		self.div_node = div_node = self.group.nodes.new('ShaderNodeMath')  # Need to keep reference so can update at runtime
		div_node.operation = 'DIVIDE'
		div_node.use_clamp = True
		div_node.inputs[1].default_value = self.N

		hsv_node = self.group.nodes.new('ShaderNodeHueSaturation')
		hsv_node.inputs['Saturation'].default_value = 1
		hsv_node.inputs['Value'].default_value = 1
		hsv_node.inputs['Color'].default_value = (1, 0, 0, 1) # Red

		self.group.links.new(self.input_node.outputs['Index'], div_node.inputs[0])
		self.group.links.new(div_node.outputs['Value'], hsv_node.inputs['Hue'])
		self.group.links.new(hsv_node.outputs['Color'], self.output_node.inputs['Color'])
		tidy_tree(self.group)

	def _add_to_shader(self, shader_node_tree):
		attr_node = shader_node_tree.nodes.new('ShaderNodeAttribute')
		attr_node.attribute_type = self.attribute_type
		attr_node.attribute_name = self.attribute_name

		# Create a group node for the Int Index -> HSV conversion
		group_node = shader_node_tree.nodes.new('ShaderNodeGroup')
		group_node.node_tree = self.group

		shader_node_tree.links.new(attr_node.outputs['Fac'], group_node.inputs['Index'])
		return group_node.outputs['Color']

	def update(self):
		self.div_node.inputs[1].default_value = self.N

	@property
	def N(self):
		return 0

class InstanceRGBAOV(AttrRGBAOV):
	"""
	Similar to InstanceIDAOV, but outputs an RGB value corresponding to the object's instance ID.
	For N instances total, samples evenly on a distribution of hues, on a colour scale of S = 1, V = 1.
	Updates N at render time by reading the scene property 'NUM_MESHES'
	"""
	attribute_type = 'OBJECT'
	attribute_name = 'instance_id'

	@property
	def N(self):
		"""Update the divisor node with the current number of instances"""
		return bpy.context.scene.get('NUM_MESHES', 0) + 1

class ClassRGBAOV(AttrRGBAOV):
	"""
	Similar to ClassIDAOV, but outputs an RGB value corresponding to the object's class ID.
	For N classes total, samples evenly on a distribution of hues, on a colour scale of S = 1, V = 1.
	Updates N at render time by reading the scene property 'MAX_CLASSES'
	"""
	attribute_type = 'OBJECT'
	attribute_name = 'class_id'

	@property
	def N(self):
		"""Update the divisor node with the current number of classes"""
		return bpy.context.scene.get('MAX_CLASSES', 0) + 1


"""Base class for all BlenderSynth objects."""
from .utils import handle_vec, SelectObjects, _euler_add, _euler_from, animatable_property
from mathutils import Vector, Euler
import numpy as np
import bpy
from typing import Union
from ..utils import types

class BsynObject:
	"""Generic class for BlenderSynth objects."""
	_object = None # corresponding blender object

	@property
	def obj(self):
		if self._object is None:
			raise ValueError("self._object not set. Ensure it is set in the Object's __init__ function.")

		return self._object

	@property
	def object(self):
		return self.obj

	@property
	def data(self):
		return self.object.data

	# def update(self):
	#   ---> not currently needed, may be needed in the future
	# 	"""On any update, run this. For most objects, this is a no-op, but for some objects,
	# 	this is necessary to update the object's state. e.g. Camera"""
	# 	return

	def _keyframe_delete(self, *args, **kwargs):
		self._object.keyframe_delete(*args, **kwargs)

	def _keyframe_insert(self, *args, **kwargs):
		self._object.keyframe_insert(*args, **kwargs)

	@property
	def _all_objects(self):
		"""List of all objects associated with this object."""
		return [self.object]

	@property
	def location(self) -> Vector:
		"""Location of object"""
		return self.obj.location

	@location.setter
	def location(self, value):
		self.set_location(value)

	@property
	def rotation_euler(self) -> Euler:
		"""Rotation in euler XYZ angles"""
		return self.obj.rotation_euler

	@rotation_euler.setter
	def rotation_euler(self, value):
		self.set_rotation_euler(value)


	@property
	def scale(self):
		return self.obj.scale

	@scale.setter
	def scale(self, scale):
		self.set_scale(scale)


	@animatable_property('location')
	def set_location(self, location: types.VectorLikeAlias):
		"""Set location of object.

		:param location: Location vector to set"""
		location = handle_vec(location, 3)

		translation = location - self.location
		with SelectObjects(self._all_objects):
			bpy.ops.transform.translate(value=translation)


	@animatable_property('rotation_euler')
	def set_rotation_euler(self, rotation: types.VectorLikeAlias):
		"""Set euler rotation of object.

		:param rotation: Rotation vector"""

		assert len(rotation) == 3, f"Rotation must be a tuple of length 3, got {len(rotation)}"
		rotation = Euler(rotation, 'XYZ')
		diff = _euler_from(self.rotation_euler, rotation)

		with SelectObjects(self._all_objects):
			for ax, val in zip('XYZ', diff):
				if val != 0:
					bpy.ops.transform.rotate(value=val, orient_axis=ax)

	@animatable_property('scale')
	def set_scale(self, scale: types.VectorLikeOrScalarAlias):
		"""Set scale of object.

		:param scale: Scale to set. Either single value or 3 long vector"""
		if isinstance(scale, (int, float)):
			scale = (scale, scale, scale)

		resize_fac = np.array(scale) / np.array(self.scale)

		with SelectObjects(self._all_objects):
			bpy.ops.transform.resize(value=resize_fac)

	def translate(self, translation):
		"""Translate object"""
		translation = handle_vec(translation, 3)
		self.location = self.location + translation

	def rotate_by(self, rotation):
		"""Add a rotation to the object. Must be in XYZ order, euler angles, radians."""
		rotation = handle_vec(rotation, 3)
		new_rotation = _euler_add(self.rotation_euler, Euler(rotation, 'XYZ'))
		self.rotation_euler = new_rotation

	def scale_by(self, scale):
		"""Scale object"""
		if isinstance(scale, (int, float)):
			scale = (scale, scale, scale)

		scale = handle_vec(scale, 3)
		self.scale *= scale

	@property
	def matrix_world(self):
		"""Return world matrix of object(s)."""
		bpy.context.evaluated_depsgraph_get()  # required to update object matrix
		return self.object.matrix_world

	@property
	def axes(self) -> np.ndarray:
		"""Return 3x3 rotation matrix (normalized) to represent axes"""
		mat = np.array(self.matrix_world)[:3, :3]
		mat = mat / np.linalg.norm(mat, axis=0)
		return mat

	def track_to(self, obj: Union['BsynObject', bpy.types.Object]):
		"""Track to object.

		:param obj: BsynObject or Blender Object to track to
		"""

		if isinstance(obj, BsynObject):
			obj = obj.obj

		constraint = self.object.constraints.new('TRACK_TO')
		constraint.target = obj
		constraint.track_axis = 'TRACK_NEGATIVE_Z'
		constraint.up_axis = 'UP_Y'

	def untrack(self):
		"""Remove track to constraint from object"""
		constraint = self.object.constraints.get('Track To')
		self.object.constraints.remove(constraint)

	def follow_path(self, path: 'Curve', zero: bool = True,
					animate: bool = True, frames: tuple = (0, 250), fracs: tuple = (0, 1)):
		"""Follow path, with optional animation setting.

		:param path: Curve object
		:param zero: If True, set camera location to (0, 0, 0) [aligns camera with path]
		:param animate: If True, animate camera along path
		:param frames: tuple of keyframes to animate at - length N
		:param fracs: tuple of fractions along path to animate at - length N
		"""

		constraint = self.object.constraints.new('FOLLOW_PATH')
		constraint.target = path.path
		constraint.forward_axis = 'TRACK_NEGATIVE_Z'
		constraint.up_axis = 'UP_Y'
		constraint.use_fixed_location = True  # ensures that offset factor is in 0-1 range

		if zero:
			self.location = (0, 0, 0)

		if animate:
			self.animate_path(frames, fracs)

		# if there are any track constraints, place this constraint first
		# so that the object is not rotated by the track constraint
		track_constraint_idx = self.object.constraints.find('Track To')
		if track_constraint_idx > -1:
			self.object.constraints.move(track_constraint_idx, track_constraint_idx + 1)

	def animate_path(self, frames: tuple = (0, 250), fracs: tuple = (0, 1)):
		"""Animate object along path.

		:param frames: tuple of keyframes to animate at - length N
		:param fracs: tuple of fractions along path to animate at - length N
		"""

		assert len(frames) == len(fracs), f"frames and fracs must be same length - got {len(frames)} and {len(fracs)}"

		for frame, frac in zip(frames, fracs):
			self.path_keyframe(frame, frac)

	def path_keyframe(self, frame: int, offset: float):
		"""Set keyframe for camera path offset

		:param frame: Frame number
		:param offset: Offset fraction (0-1)
		"""
		constraint = self.object.constraints.get('Follow Path')
		if constraint is None:
			raise ValueError(f"Object `{self.__class__}` does not have a 'Follow Path' constraint")

		constraint.offset_factor = offset
		constraint.keyframe_insert(data_path='offset_factor', frame=frame)


import bpy
import numpy as np
from ..utils import get_node_by_name
import os
import shutil
from ..render import render, render_depth
from .node_group import CompositorNodeGroup
from ..aov import AOV
from ..mesh import Mesh
from .mask_overlay import MaskOverlay
from .visuals import DepthVis
from .image_overlay import KeypointsOverlay, BoundingBoxOverlay, AlphaImageOverlay, AxesOverlay


from typing import Union, List
from ...utils.node_arranger import tidy_tree

from ..world import world

# Mapping of file formats to extensions
format_to_extension = {
	'BMP': '.bmp',
	'IRIS': '.rgb',
	'PNG': '.png',
	'JPEG': '.jpg',
	'JPEG2000': '.jp2',
	'TARGA': '.tga',
	'TARGA_RAW': '.tga',
	'CINEON': '.cin',
	'DPX': '.dpx',
	'OPEN_EXR_MULTILAYER': '.exr',
	'OPEN_EXR': '.exr',
	'HDR': '.hdr',
	'TIFF': '.tif',
	# Add more formats if needed
}


def get_badfname(fname, N=100):
	"""Search for filename in the format
	<main_fname><i:04d>.<ext>
	where i is the frame number. if no file found for i < N, raise error.
	otherwise, return found file
	"""
	f, ext = os.path.splitext(fname)
	for i in range(N):
		fname = f + f"{i:04d}" + ext
		if os.path.isfile(fname):
			return fname

	raise FileNotFoundError(f"File {fname} not found")


def remove_ext(fname):
	return os.path.splitext(fname)[0]  # remove extension if given


class Compositor:
	"""Compositor output - for handling file outputs, and managing Compositor node tree"""

	def __init__(self, view_layer='ViewLayer', background_color:tuple=None,
				 rgb_color_space:str='Filmic sRGB'):
		"""
		:param view_layer: Name of View Layer to render
		:param background_color: If given, RGB[A] tuple in range [0-1], will overwrite World background with solid color (while retaining lighting effects).
		:param rgb_color_space: Color transform for RGB only.
		"""
		# Create compositor node tree
		bpy.context.scene.use_nodes = True
		self.node_tree = bpy.context.scene.node_tree
		# self.file_output_nodes = []

		self.view_layer = view_layer#

		self.file_output_nodes = {}  # Mapping of output name to FileOutputNode
		self.mask_nodes = {}  # Mapping of mask pass index to CompositorNodeGroup
		self.overlays = {}
		self.aovs = []  # List of AOVs (used to update before rendering)

		# We set view transform to 'Raw' to avoid any gamma correction to all non-Image layers
		bpy.context.scene.view_settings.view_transform = 'Raw'

		# Socket to be used as RGB input for anything. Defined separately in case of applying overlays (e.g. background colour)
		self._rgb_socket = get_node_by_name(self.node_tree, 'Render Layers').outputs['Image']
		self._set_rgb_color_space(rgb_color_space)
		if background_color is not None:
			self._set_background_color(background_color)

	def tidy_tree(self):
		"""Tidy up node tree"""
		tidy_tree(self.node_tree)

	@property
	def render_layers_node(self):
		return get_node_by_name(self.node_tree, 'Render Layers')

	def _get_render_layer_output(self, key:str):
		"""Get output socket from Render Layers node"""
		if key == 'Image': # special case
			return self._rgb_socket
		else:
			return self.render_layers_node.outputs[key]

	def get_mask(self, index, input_rgb: Union[str, CompositorNodeGroup], anti_aliasing=False) -> CompositorNodeGroup:
		"""Get mask node from pass index. If not found, create new mask node"""
		bpy.context.scene.view_layers[self.view_layer].use_pass_object_index = True  # Make sure object index is enabled

		if index not in self.mask_nodes:

			if isinstance(input_rgb, str):
				ip_node = self._get_render_layer_output(input_rgb)

			elif isinstance(input_rgb, CompositorNodeGroup):
				ip_node = input_rgb.outputs['Image']

			else:
				raise TypeError(f"input_rgb must be str or CompositorNodeGroup, got {type(input_rgb)}")

			dtype = 'Float' if isinstance(ip_node, bpy.types.NodeSocketFloat) else 'Color'

			cng = MaskOverlay(f"Mask - ID: {index} - Input {input_rgb}",
							  self.node_tree, index=index, dtype=dtype,
							  use_antialiasing=anti_aliasing)

			self.node_tree.links.new(ip_node, cng.input('Image'))
			self.node_tree.links.new(self._get_render_layer_output('IndexOB'), cng.input('IndexOB'))
			self.mask_nodes[index] = cng

		self.tidy_tree()
		return self.mask_nodes[index]

	def get_bounding_box_visual(self, col=(0., 0., 255., 255.), thickness=5) -> BoundingBoxOverlay:
		"""
		return a CompositorNodeGroup,
		which will render the bounding boxes of the objects

		:param col: (3,) or (N, 3) Color(s) of bounding box(es) [in BGR]
		:param thickness: (,) or (N,) Thickness(es) of bounding box(es)
		"""

		cng = BoundingBoxOverlay(f"Bounding Box Visual", self.node_tree, col=col, thickness=thickness)
		self.node_tree.links.new(self._get_render_layer_output('Image'), cng.input('Image'))

		if 'BBox' in self.overlays:
			raise ValueError("Only allowed one BBox overlay (it can contain multiple objects).")

		self.overlays['BBox'] = cng

		self.tidy_tree()
		return cng

	def get_keypoints_visual(self, marker:str='x', color:tuple=(0, 0, 255), size:int=5,
							 thickness:int=2) -> KeypointsOverlay:
		"""
		Initialize a keypoints overlay node.

		:param marker: Marker type, either [c/circle], [s/square], [t/triangle] or [x]. Default 'x'
		:param size: Size of marker. Default 5
		:param color: Color of marker, RGB or RGBA, default (0, 0, 255) (red)
		:param thickness: Thickness of marker. Default 2
		"""

		cng = KeypointsOverlay(f"Keypoints Visual", self.node_tree, marker=marker, color=color,
							   size=size, thickness=thickness)
		self.node_tree.links.new(self._get_render_layer_output('Image'), cng.input('Image'))

		if 'Keypoints' in self.overlays:
			raise ValueError("Only allowed one Keypoints overlay.")

		self.overlays['Keypoints'] = cng

		self.tidy_tree()
		return cng

	def get_axes_visual(self, size:int=1, thickness:int=2) -> AxesOverlay:
		"""
		Initialize an axes overlay node.

		:param size: Size of axes. Default 100
		:param thickness: Thickness of axes. Default 2
		"""

		cng = AxesOverlay(f"Axes Visual", self.node_tree,
							   size=size, thickness=thickness)
		self.node_tree.links.new(self._get_render_layer_output('Image'), cng.input('Image'))

		if 'Axes' in self.overlays:
			raise ValueError("Only allowed one Axes overlay.")

		self.overlays['Axes'] = cng
		self.tidy_tree()
		return cng

	def stack_visuals(self, *visuals: AlphaImageOverlay) -> AlphaImageOverlay:
		"""Given a series of image overlays, stack them and return to be used as a single output node.

		:param *visuals: Stack of overlays to add."""

		if len(visuals) < 2:
			raise ValueError("Requires at least 2 visuals to stack")

		# No need to store these overlays separately in self.overlays, but need to check they're all present
		for overlay in visuals:
			if overlay not in self.overlays.values():
				raise ValueError(f"Visual {overlay} not found in Compositor. Make sure it was obtained via the Compositor.")

		# Stack the output of the previous to the input of the next
		for va, vb in zip(visuals, visuals[1:]):
			self.node_tree.links.new(va.output('Image'), vb.input('Image'))

		return visuals[-1]

	def get_depth_visual(self, max_depth=1, col=(255, 255, 255)):
		"""Get depth visual, which normalizes depth values so max_depth = col,
		and any values below that are depth/max_depth * col.

		Col = 0-255 RGB or RGBA"""

		if 'Depth' not in self.render_layers_node.outputs:
			render_depth()

		# convert col to 0-1, RGBA
		col = ([i/255 for i in col] + [1])[:4]

		cng = DepthVis(self.node_tree, max_depth=max_depth, col=col)
		self.node_tree.links.new(self._get_render_layer_output('Depth'), cng.input('Depth'))

		self.tidy_tree()
		return cng

	def define_output(self, input_data: Union[str, CompositorNodeGroup, AOV], directory, file_name=None, mode='image',
					  file_format='PNG', color_mode='RGBA', jpeg_quality=90,
					  png_compression=15, color_depth='8', EXR_color_depth='32',
					  name=None):
		"""Add a connection between a valid render output, and a file output node.
		Supports changing view output.

		This should only be called once per output (NOT inside a loop).
		Inside the loop, only call `update_filename` or `update_directory`

		:mode: if 'image', export in sRGB color space. If 'data', export in raw linear color space

		:input_data: Can take one of three forms:
			- `string`, will get the input_data from that key in the render_layers_node
			- `CompositorNodeGroup`, will use that node as input
			- `AOV`, will use that AOV as input (storing AOV)

		:name: Name of output. If not given, will take the str representation of input_data
		"""

		if isinstance(input_data, AOV):
			self.aovs.append(input_data)
			input_data = input_data.name  # name is sufficient to pull from render_layers_node

		assert mode in ['image', 'data'], f"mode must be 'image' or 'data', got {mode}"
		assert file_format in format_to_extension, f"File format `{file_format}` not supported. Options are: {list(format_to_extension.keys())}"

		if name is None:
			name = str(input_data)

		node_name = f"File Output {name}"

		# check node doesn't exist
		if name in self.file_output_nodes:
			raise ValueError(f"File output `{name}` already exists. Only call define_output once per output type.")


		if file_name is None: # if fname is not given, use input_name
			file_name = name

		file_name = remove_ext(file_name)
		directory = os.path.abspath(directory)  # make sure directory is absolute
		os.makedirs(directory, exist_ok=True)

		# Create new 'File Output' node in compositor
		node = self.node_tree.nodes.new('CompositorNodeOutputFile')
		node.name = node_name

		if isinstance(input_data, str):
			self.node_tree.links.new(self._get_render_layer_output(input_data), node.inputs['Image'])

		elif isinstance(input_data, CompositorNodeGroup):  # add overlay in between
			self.node_tree.links.new(input_data.outputs[0], node.inputs['Image'])

		else:
			raise NotImplementedError(
				f"input_data must be either str or CompositorNodeGroup, got {type(input_data)}")

		# Set file output node properties
		node.label = node_name
		node.base_path = directory
		node.file_slots[0].path = file_name

		if mode == 'data':
			node.format.color_management = 'OVERRIDE'
			node.format.display_settings.display_device = 'None'

		# File format kwargs
		node.format.file_format = file_format
		node.format.color_mode = color_mode
		node.format.quality = jpeg_quality
		node.format.compression = png_compression
		node.format.color_depth = color_depth if file_format != 'OPEN_EXR' else EXR_color_depth

		self.file_output_nodes[name] = node

		self.tidy_tree()
		return name

	def update_filename(self, key: str, fname: str):
		"""Reassign the filename (not directory) for a given file output node.

		:param key: key of output, as given in `define_output`
		:param fname: new filename, without extension"""
		fname = remove_ext(fname)
		node = self.file_output_nodes[str(key)]
		node.file_slots[0].path = fname

	def update_all_filenames(self, fname:str):
		"""Reassign all filenames (not directories) for all file output nodes.

		:param fname: new filename, without extension"""
		fname = remove_ext(fname)
		for node in self.file_output_nodes.values():
			node.file_slots[0].path = fname

	def update_directory(self, key, directory):
		"""Reassign the directory for a given file output node"""
		node = self.file_output_nodes[str(key)]
		node.base_path = directory

	def fix_namings(self):
		"""After rendering,
		File Output node has property that frame number gets added to filename.
		Fix that here"""

		for node in self.file_output_nodes.values():
			# get expected file name and extension
			ext = format_to_extension[node.format.file_format]
			target_file_name = os.path.join(node.base_path, node.file_slots[0].path + ext)
			bad_file_name = get_badfname(target_file_name)
			shutil.move(
				bad_file_name,
				target_file_name
			)

	def update_aovs(self):
		"""Update any AOVs that are connected to the render layers node"""
		for aov in self.aovs:
			aov.update()

	def render(self, camera=None, scene=None, overlay_kwargs=None,
			   animation=False, frame_start=0, frame_end=250):
		"""Render the scene"""
		if overlay_kwargs is None:
			overlay_kwargs = {}

		if scene is None:
			scene = bpy.context.scene

		if camera is None:
			camera = scene.camera

		for k in overlay_kwargs.keys():
			assert k in self.overlays, f"overlay_kwarg {k} not in overlays: {[*self.overlays.keys()]}."

		for oname, overlay in self.overlays.items():
			args = overlay_kwargs.get(oname, {})
			if isinstance(args, dict):
				overlay.update(camera=camera, scene=scene, **args)  # multi kwargs
			else:
				overlay.update(args, camera=camera, scene=scene)  # single arg


		self.update_aovs()

		if animation:
			scene.frame_start = frame_start
			scene.frame_end = frame_end

		render(animation=animation)

		if not animation:
			self.fix_namings()

	def _set_rgb_color_space(self, color_space:str='Filmic sRGB'):
		"""Color spaces are all handled manually within compositor (so that we keep
		AOVs in raw space). So set the color space for RGB socket here."""

		color_space_node = self.node_tree.nodes.new('CompositorNodeConvertColorSpace')
		color_space_node.from_color_space = 'Linear'
		color_space_node.to_color_space = color_space

		self.node_tree.links.new(self._rgb_socket, color_space_node.inputs[0])
		self._rgb_socket = color_space_node.outputs[0]

	def _set_background_color(self, color:tuple=(0, 0, 0)):
		"""Set a solid background color, instead of transparent.
		Will remove the visuals of existing world background (but not the lighting effects).

		:param color: RGBA color, in range [0, 1]
		"""

		world.set_transparent()

		rgba = color
		if len(rgba) == 3:
			rgba = (*rgba, 1)

		rgb_node = self.node_tree.nodes.new('CompositorNodeRGB')
		rgb_node.outputs[0].default_value = rgba

		mix_node = self.node_tree.nodes.new('CompositorNodeMixRGB')

		self.node_tree.links.new(self._rgb_socket, mix_node.inputs[2])
		self.node_tree.links.new(rgb_node.outputs[0], mix_node.inputs[1])

		self.node_tree.links.new(self._get_render_layer_output('Alpha'), mix_node.inputs['Fac'])
		self._rgb_socket = mix_node.outputs['Image']


"""Custom compositor node group"""
import bpy
from ...utils.node_arranger import tidy_tree

class CompositorNodeGroup:
	def __init__(self, name, node_tree):
		self.name = name
		self.node_tree = node_tree
		self.group = bpy.data.node_groups.new(type='CompositorNodeTree', name=name)

		self.gn = group_node = node_tree.nodes.new("CompositorNodeGroup")
		group_node.node_tree = self.group

		self.input_node = self.group.nodes.new("NodeGroupInput")
		self.output_node = self.group.nodes.new("NodeGroupOutput")

	def tidy(self):
		tidy_tree(self.group)


	@property
	def inputs(self):
		return self.gn.inputs

	@property
	def outputs(self):
		return self.gn.outputs

	def input(self, name):
		return self.inputs[name]

	def output(self, name):
		return self.outputs[name]

	def __str__(self):
		return f"CompositorNodeGroup({self.name})"

	def update(self, camera=None, scene=None):
		pass


import bpy
from .utils import project_points
import mathutils
import numpy as np
from ..blender.mesh import Mesh
from typing import List

BBOX_FMTS = ['x1y1x2y2', 'xywh']

def bounding_box(object: Mesh, camera: bpy.types.Camera = None,
				 scene:bpy.types.Scene=None,
				 return_fmt:str='x1y1x2y2', normalized:bool=False, invert_y:bool=True) -> tuple:
	"""Get the bounding box of an object in camera space.
	Achieve this by projecting all vertices of the object to the image plane,
	and taking the min/max of the resulting coordinates.

	:param object: Mesh
	:param camera: bpy.types.Camera (if None, use bpy.context.scene.camera)
	:param scene: bpy.types.Scene (if None, use bpy.context.scene)
	:param return_fmt: one of ['x1y1x2y2', 'xywh']
	:param normalized: if True, return normalized coordinates (0-1) instead of pixel coordinates
	:param invert_y: if True, y is measured from the top of the image, otherwise from the bottom (Blender measures from bottom)
	"""

	if camera is None:
		camera = bpy.context.scene.camera

	if scene is None:
		scene = bpy.context.scene

	verts = object.get_all_vertices('WORLD')
	coords_2d = project_points(verts, scene, camera, invert_y=invert_y)

	if normalized:
		coords_2d[:, 0] /= scene.render.resolution_x
		coords_2d[:, 1] /= scene.render.resolution_y

	xmin, ymin = coords_2d.min(axis=0)
	xmax, ymax = coords_2d.max(axis=0)


	if return_fmt == 'x1y1x2y2':
		return xmin, ymin, xmax, ymax

	elif return_fmt == 'xywh':
		return xmin, ymin, xmax - xmin, ymax - ymin

	else:
		raise ValueError(f'Invalid return_fmt: {return_fmt}. Must be one of {BBOX_FMTS}')

def bounding_boxes(objects: List[Mesh], camera: bpy.types.Camera = None,
				 scene:bpy.types.Scene=None,
				 return_fmt:str='x1y1x2y2', normalized:bool=False, invert_y:bool=True) -> List[tuple]:
	"""Run `bounding_box` for multiple objects

	:param objects: List of Mesh objects
	:param camera: bpy.types.Camera (if None, use bpy.context.scene.camera)
	:param scene: bpy.types.Scene (if None, use bpy.context.scene)
	:param return_fmt: one of ['x1y1x2y2', 'xywh']
	:param normalized: if True, return normalized coordinates (0-1) instead of pixel coordinates
	:param invert_y: if True, y is measured from the top of the image, otherwise from the bottom (Blender measures from bottom)
	"""

	return [bounding_box(obj, camera=camera, scene=scene, return_fmt=return_fmt,
						   normalized=normalized, invert_y=invert_y) for obj in objects]


import bpy

ENGINES = ['BLENDER_EEVEE', 'BLENDER_WORKBENCH', 'CYCLES']
"""List of valid render engines"""


def render(animation:bool=False):
	"""Render the scene

	:param animation: If True, render the animation. Otherwise, render a single frame."""
	if animation:
		bpy.ops.render.render(animation=True)
	else:
		bpy.ops.render.render(write_still=True)


def set_engine(engine: str):
	"""Set the render engine

	:param engine: The render engine to use. See :data:`~blendersynth.blender.render.ENGINES` for valid options."""
	assert engine in ENGINES, "Invalid render engine"
	bpy.context.scene.render.engine = engine


def set_resolution(x: int, y: int):
	"""Set the render resolution

	:param x: The width of the image in pixels
	:param y: The height of the image in pixels"""
	bpy.context.scene.render.resolution_x = x
	bpy.context.scene.render.resolution_y = y


def set_cycles_samples(samples: int):
	"""Set the number of samples for the Cycles renderer"""
	assert bpy.context.scene.render.engine == 'CYCLES', "Cycles must be the active render engine"
	bpy.context.scene.cycles.samples = samples


def render_depth():
	"""Enable the depth pass of the renderer"""
	bpy.context.view_layer.use_pass_z = True  # enable depth pass


def set_transparent(scene=None):
	"""Set the background of the scene to transparent

	:param scene: The scene to set the background of. If None, use the active scene.
	"""
	if scene is None:
		scene = bpy.context.scene
	scene.render.film_transparent = True



import json
import os
from typing import Union


def save_label(data: Union[dict, list], pth: str):
	"""Save a label to a file.

	:param data: The data to save.
	:param pth: The path to save the data to."""

	os.makedirs(os.path.dirname(pth), exist_ok=True)
	with open(pth, 'w') as outfile:
		json.dump(data, outfile)



import bpy
from .utils import project_points
import numpy as np

def project_keypoints(points3d: np.ndarray, scene:bpy.types.Scene=None, camera:bpy.types.Camera=None):
	"""Project 3D world points to 2D image coordinates.

	Projects using :func:`blendersynth.annotations.utils.project_points`.

	:param points3d: Nx3 array of 3D points
	:param scene: scene to use (if None, `use bpy.context.scene`)
	:param camera: camera to project through (if None, use `bpy.context.scene.camera`)
	"""

	if scene is None:
		scene = bpy.context.scene

	if camera is None:
		camera = scene.camera

	# project points
	return project_points(points3d, scene, camera).tolist()



import appdirs
import os
import configparser
import shutil
import subprocess
import warnings

appname = "blendersynth"
appauthor = "BlenderSynth"
config_dir = appdirs.user_config_dir(appname, appauthor)
config_file = os.path.join(config_dir, "config.ini")


def is_blender_in_path() -> bool:
	"""Check if blender is in the PATH (i.e. can be run from the command line with 'blender')"""
	return shutil.which("blender") is not None


def find_blender_python(blender_path: str) -> str:
	"""Given a blender executable, find the python interpreter it uses

	:param blender_path: path to blender executable"""

	if read_from_config('BLENDER_PYTHON_PATH') is not None:
		return read_from_config('BLENDER_PYTHON_PATH')

	# get blender to run script to find python path
	try:
		targ_script = os.path.abspath(os.path.join(os.path.dirname(__file__), 'blender_python_path.py'))
		output = subprocess.check_output([blender_path, "--background", "--python", targ_script])
	except subprocess.CalledProcessError as e:
		raise Exception(f"Issues with finding blender python path. Error: {e}")

	# read output to find python path
	out = None
	for l in output.decode('utf-8').split('\n'):
		if "PYTHON INTERPRETER" in l:
			out = l.split(": ")[1].strip()
			break

	if out is not None:
		write_to_config('BLENDER_PYTHON_PATH', out)
		return out

	raise Exception("Could not find Python interpreter for Blender.")


def validate_blender_path(blender_path: str) -> bool:
	"""Check if `blender_path` is a valid blender executable.

	:param blender_path: path to blender executable"""
	if os.access(blender_path, os.X_OK):
		return True

	return False


def get_blender_path(_blender_path: str = None) -> str:
	"""Get a valid blender executable path.

	To achieve this, checks in the following locations:

	1) Input argument `_blender_path` (if given)

	2) Config file stored in user's `appdirs.user_config_dir`

	3) Environment variable `BLENDER_PATH`

	4) Blender in PATH

	5) Ask user for path

	It will end as soon as a valid path is found. If the input path is not valid, it will raise an error.

	:param _blender_path: path to blender executable
	"""

	cfg_result = read_from_config('BLENDER_PATH')

	if _blender_path is not None:

		if validate_blender_path(_blender_path):
			blender_path = _blender_path

		elif validate_blender_path(_blender_path + '.exe'):  # add .exe if not there (helps os.access checking)
			blender_path = _blender_path + '.exe'

		else:
			raise ValueError(f"Provided Blender path, {_blender_path}, is not executable.")

	elif cfg_result is not None:
		blender_path = cfg_result

	elif os.environ.get('BLENDER_PATH') is not None:
		blender_path = os.environ.get('BLENDER_PATH')

	elif is_blender_in_path():
		blender_path = shutil.which("blender")

	else:
		blender_path = input(
			"Blender not found in PATH or Environment Variable.\nPlease provide path to blender executable: ")

	blender_path = os.path.abspath(blender_path)  # make sure it's absolute path

	if not validate_blender_path(blender_path):
		if validate_blender_path(blender_path + '.exe'):
			blender_path += '.exe'

		else:
			raise ValueError(f"Provided Blender path, {blender_path}, is not valid as a Blender executable.")

	write_to_config('BLENDER_PATH', blender_path)
	return blender_path


def write_to_config(key, value, section='BLENDER_SETUP'):
	"""Load config, and write key value pair to cfg[section]"""
	config = configparser.ConfigParser()

	if os.path.exists(config_file):
		config.read(config_file)

	if section not in config:
		config[section] = {}

	config[section][key] = value

	with open(config_file, 'w') as configfile:
		config.write(configfile)


def read_from_config(key, section='BLENDER_SETUP'):
	"""Load config, and read value from cfg[section]"""
	config = configparser.ConfigParser()

	if os.path.exists(config_file):
		config.read(config_file)

	if section not in config:
		return None

	if key not in config[section]:
		return None

	return config[section][key]


def remove_from_config(key, section='BLENDER_SETUP'):
	"""Load config, and remove key from cfg[section]"""
	config = configparser.ConfigParser()

	if os.path.exists(config_file):
		config.read(config_file)

	if section not in config:
		return None

	if key not in config[section]:
		return None

	del config[section][key]

	with open(config_file, 'w') as configfile:
		config.write(configfile)


def remove_config():
	"""Remove config file"""
	if os.path.exists(config_file):
		os.remove(config_file)
